Set Implicit Arguments.
Require Import MyTactics.
Require Import MyLists.
Require Import DeBruijn.
Require Import SourceSyntax.
Require Import TargetSyntax.
Require Import SourceSemantics.
Require Import TargetSemantics.
Require Import EncodingBricks.
Require Import Encoding.

(* Lots of facts about closed terms, in the source and target languages. *)

(* I must recognize that my decision to define well-scoped terms at depth [k]
   in an indirect manner -- see the definition of [well_scoped_lift] -- was
   not smart. The idea was to avoid duplicating the definition of terms, but
   I ended up having to prove a lot of construction and deconstruction lemmas
   about well-scopedness and closedness, so the hassle was greater in the end. *)

(* ------------------------------------------------------------------------- *)

(* Closedness is hereditary -- if a term is closed, then its subterms are
   closed. *)

Ltac well_scoped_hereditary_tac :=
  unfold well_scoped_lift;
  intros;
  match goal with
  | h1: forall j, j >= ?k -> _, h2: ?j >= ?k |- _ =>
    specializes h1 h2; simpl in h1
  end;
  congruence.

Ltac closed_hereditary_tac :=
  unfold closed_sterm, closed_tterm, closed, well_scoped_lift;
  intros;
  match goal with
  | h1: forall k j, j >= k -> _, h2: ?j >= ?k |- _ =>
    specializes h1 h2; simpl in h1
  end;
  congruence.

Lemma well_scoped_SAbs:
  forall t k,
  well_scoped_lift lift_sterm k (SAbs t) ->
  well_scoped_lift lift_sterm (S k) t.
Proof.
  unfold well_scoped_lift.
  intros ? ? h1 j h2. destruct j. false; omega.
  specializes h1 j. simpl in h1.
  forwards: h1. omega. congruence.
Qed.

Lemma well_scoped_SApp_1:
  forall t1 t2 k,
  well_scoped_lift lift_sterm k (SApp t1 t2) ->
  well_scoped_lift lift_sterm k t1.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_SApp_2:
  forall t1 t2 k,
  well_scoped_lift lift_sterm k (SApp t1 t2) ->
  well_scoped_lift lift_sterm k t2.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_SPair_1:
  forall t1 t2 k,
  well_scoped_lift lift_sterm k (SPair t1 t2) ->
  well_scoped_lift lift_sterm k t1.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_SPair_2:
  forall t1 t2 k,
  well_scoped_lift lift_sterm k (SPair t1 t2) ->
  well_scoped_lift lift_sterm k t2.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_SProjLeft:
  forall t k,
  well_scoped_lift lift_sterm k (SProjLeft t) ->
  well_scoped_lift lift_sterm k t.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_SProjRight:
  forall t k,
  well_scoped_lift lift_sterm k (SProjRight t) ->
  well_scoped_lift lift_sterm k t.
Proof.
  well_scoped_hereditary_tac.
Qed.

Hint Resolve well_scoped_SAbs well_scoped_SApp_1 well_scoped_SApp_2
well_scoped_SPair_1 well_scoped_SPair_2 well_scoped_SProjLeft
well_scoped_SProjRight : ws.

Ltac corollary :=
  unfold closed_sterm, closed_tterm, closed;
  eauto 3 with ws.

Lemma closed_SAbs:
  forall t,
  closed_sterm (SAbs t) ->
  well_scoped_lift lift_sterm 1 t.
Proof.
  corollary.
Qed.

Lemma closed_SApp_1:
  forall t1 t2,
  closed_sterm (SApp t1 t2) ->
  closed_sterm t1.
Proof.
  corollary.
Qed.

Lemma closed_SApp_2:
  forall t1 t2,
  closed_sterm (SApp t1 t2) ->
  closed_sterm t2.
Proof.
  corollary.
Qed.

Lemma closed_SPair_1:
  forall t1 t2,
  closed_sterm (SPair t1 t2) ->
  closed_sterm t1.
Proof.
  corollary.
Qed.

Lemma closed_SPair_2:
  forall t1 t2,
  closed_sterm (SPair t1 t2) ->
  closed_sterm t2.
Proof.
  corollary.
Qed.

Lemma closed_SProjLeft:
  forall t,
  closed_sterm (SProjLeft t) ->
  closed_sterm t.
Proof.
  corollary.
Qed.

Lemma closed_SProjRight:
  forall t,
  closed_sterm (SProjRight t) ->
  closed_sterm t.
Proof.
  corollary.
Qed.

Lemma well_scoped_TAbs:
  forall t k,
  well_scoped_lift lift_tterm k (TAbs t) ->
  well_scoped_lift lift_tterm (S k) t.
Proof.
  unfold well_scoped_lift.
  intros ? ? h1 j h2. destruct j. false; omega.
  specializes h1 j. simpl in h1.
  forwards: h1. omega. congruence.
Qed.

Lemma well_scoped_TApp_1:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k (TApp t1 t2) ->
  well_scoped_lift lift_tterm k t1.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_TApp_2:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k (TApp t1 t2) ->
  well_scoped_lift lift_tterm k t2.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_TPair_1:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k (TPair t1 t2) ->
  well_scoped_lift lift_tterm k t1.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_TPair_2:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k (TPair t1 t2) ->
  well_scoped_lift lift_tterm k t2.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_TLetPair_1:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k (TLetPair t1 t2) ->
  well_scoped_lift lift_tterm k t1.
Proof.
  well_scoped_hereditary_tac.
Qed.

Lemma well_scoped_TLetPair_2:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k (TLetPair t1 t2) ->
  well_scoped_lift lift_tterm (S (S k)) t2.
Proof.
  unfold well_scoped_lift.
  intros ? ? ? h1 j h2. destruct j. false; omega. destruct j. false; omega.
  specializes h1 j. simpl in h1.
  forwards: h1. omega. congruence.
Qed.

Hint Resolve well_scoped_TAbs well_scoped_TApp_1 well_scoped_TApp_2
well_scoped_TPair_1 well_scoped_TPair_2 well_scoped_TLetPair_2 : ws.

Lemma closed_TAbs:
  forall t,
  closed_tterm (TAbs t) ->
  well_scoped_lift lift_tterm 1 t.
Proof.
  corollary.
Qed.

Lemma closed_TApp_1:
  forall t1 t2,
  closed_tterm (TApp t1 t2) ->
  closed_tterm t1.
Proof.
  closed_hereditary_tac.
Qed.

Lemma closed_TApp_2:
  forall t1 t2,
  closed_tterm (TApp t1 t2) ->
  closed_tterm t2.
Proof.
  closed_hereditary_tac.
Qed.

Lemma closed_TPair_1:
  forall t1 t2,
  closed_tterm (TPair t1 t2) ->
  closed_tterm t1.
Proof.
  closed_hereditary_tac.
Qed.

Lemma closed_TPair_2:
  forall t1 t2,
  closed_tterm (TPair t1 t2) ->
  closed_tterm t2.
Proof.
  closed_hereditary_tac.
Qed.

Lemma closed_TLetPair_1:
  forall t1 t2,
  closed_tterm (TLetPair t1 t2) ->
  closed_tterm t1.
Proof.
  closed_hereditary_tac.
Qed.

Lemma closed_TLetPair_2:
  forall t1 t2,
  closed_tterm (TLetPair t1 t2) ->
  well_scoped_lift lift_tterm 2 t2.
Proof.
  corollary.
Qed.

Ltac closed_hereditary :=
  match goal with
  | h: closed_sterm (SAbs _) |- _ =>
      unpack (closed_SAbs h); clear h
  | h: closed_sterm (SApp _ _) |- _ =>
      unpack (closed_SApp_1 h);
      unpack (closed_SApp_2 h);
      clear h;
      closed_hereditary
  | h: closed_sterm (SPair _ _) |- _ =>
      unpack (closed_SPair_1 h);
      unpack (closed_SPair_2 h);
      clear h;
      closed_hereditary
  | h: closed_sterm (SProjLeft _) |- _ =>
      unpack (closed_SProjLeft h);
      clear h;
      closed_hereditary
  | h: closed_sterm (SProjRight _) |- _ =>
      unpack (closed_SProjRight h);
      clear h;
      closed_hereditary
  | h: closed_ectx (EBind _ _) |- _ =>
      inversion h;
      clear h;
      closed_hereditary
  | h: well_scoped_lift lift_sterm _ (SAbs _) |- _ =>
      unpack (well_scoped_SAbs h); clear h
  | h: well_scoped_lift lift_sterm _ (SApp _ _) |- _ =>
      unpack (well_scoped_SApp_1 h);
      unpack (well_scoped_SApp_2 h);
      clear h;
      closed_hereditary
  | h: well_scoped_lift lift_sterm _ (SPair _ _) |- _ =>
      unpack (well_scoped_SPair_1 h);
      unpack (well_scoped_SPair_2 h);
      clear h;
      closed_hereditary
  | h: well_scoped_lift lift_sterm _ (SProjLeft _) |- _ =>
      unpack (well_scoped_SProjLeft h);
      clear h;
      closed_hereditary
  | h: well_scoped_lift lift_sterm _ (SProjRight _) |- _ =>
      unpack (well_scoped_SProjRight h);
      clear h;
      closed_hereditary
  | h: closed_tterm (TAbs _) |- _ =>
      unpack (closed_TAbs h); clear h
  | h: closed_tterm (TApp _ _) |- _ =>
      unpack (closed_TApp_1 h);
      unpack (closed_TApp_2 h);
      clear h;
      closed_hereditary
  | h: closed_tterm (TPair _ _) |- _ =>
      unpack (closed_TPair_1 h);
      unpack (closed_TPair_2 h);
      clear h;
      closed_hereditary
  | h: closed_tterm (TLetPair _ _) |- _ =>
      unpack (closed_TLetPair_1 h);
      unpack (closed_TLetPair_2 h);
      clear h;
      closed_hereditary
  | h: well_scoped_lift lift_tterm _ (TAbs _) |- _ =>
      unpack (well_scoped_TAbs h); clear h
  | h: well_scoped_lift lift_tterm _ (TApp _ _) |- _ =>
      unpack (well_scoped_TApp_1 h);
      unpack (well_scoped_TApp_2 h);
      clear h;
      closed_hereditary
  | h: well_scoped_lift lift_tterm _ (TPair _ _) |- _ =>
      unpack (well_scoped_TPair_1 h);
      unpack (well_scoped_TPair_2 h);
      clear h;
      closed_hereditary
  | h: well_scoped_lift lift_tterm _ (TLetPair _ _) |- _ =>
      unpack (well_scoped_TLetPair_1 h);
      unpack (well_scoped_TLetPair_2 h);
      clear h;
      closed_hereditary
  | _ =>
      idtac
  end.

(* ------------------------------------------------------------------------- *)

(* Closedness is contagious -- if a term's subterms are closed, then the term
   itself is closed. *)

Ltac closed_contagious :=
  unfold closed_sterm, closed_tterm, closed, well_scoped_lift in *;
  simpl; intros; f_equal; eauto.

Lemma well_scoped_up_SReturn:
  forall k,
  well_scoped_lift lift_sterm k SReturn.
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_SAbs:
  forall t k,
  well_scoped_lift lift_sterm (S k) t ->
  well_scoped_lift lift_sterm k (SAbs t).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_SApp:
  forall t1 t2 k,
  well_scoped_lift lift_sterm k t1 ->
  well_scoped_lift lift_sterm k t2 ->
  well_scoped_lift lift_sterm k (SApp t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_SPair:
  forall t1 t2 k,
  well_scoped_lift lift_sterm k t1 ->
  well_scoped_lift lift_sterm k t2 ->
  well_scoped_lift lift_sterm k (SPair t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_SProjLeft:
  forall t k,
  well_scoped_lift lift_sterm k t ->
  well_scoped_lift lift_sterm k (SProjLeft t).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_SProjRight:
  forall t k,
  well_scoped_lift lift_sterm k t ->
  well_scoped_lift lift_sterm k (SProjRight t).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_SUnit:
  forall k,
  well_scoped_lift lift_sterm k SUnit.
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_SLoc:
  forall l k,
  well_scoped_lift lift_sterm k (SLoc l).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_SReturn:
  closed_sterm SReturn.
Proof.
  closed_contagious.
Qed.

Lemma closed_up_SApp:
  forall t1 t2,
  closed_sterm t1 ->
  closed_sterm t2 ->
  closed_sterm (SApp t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_SPair:
  forall t1 t2,
  closed_sterm t1 ->
  closed_sterm t2 ->
  closed_sterm (SPair t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_SProjLeft:
  forall t,
  closed_sterm t ->
  closed_sterm (SProjLeft t).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_SProjRight:
  forall t,
  closed_sterm t ->
  closed_sterm (SProjRight t).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_SUnit:
  closed_sterm SUnit.
Proof.
  closed_contagious.
Qed.

Lemma closed_up_SLoc:
  forall l,
  closed_sterm (SLoc l).
Proof.
  closed_contagious.
Qed.

Hint Resolve well_scoped_up_SReturn well_scoped_up_SAbs
well_scoped_up_SApp well_scoped_up_SPair well_scoped_up_SProjLeft
well_scoped_up_SProjRight well_scoped_up_SUnit well_scoped_up_SLoc :
ws.

Hint Resolve closed_up_SReturn closed_up_SApp closed_up_SPair
closed_up_SProjLeft closed_up_SProjRight closed_up_SUnit
closed_up_SLoc.

(* ------------------------------------------------------------------------- *)

(* Lifting a closed term yields a closed term. *)

Lemma closed_lift_closed_tterm:
  forall k t,
  closed_tterm t ->
  closed_tterm (lift_tterm k t).
Proof.
  intros. rewrite lift_closed_tterm; eauto.
Qed.

Hint Resolve closed_lift_closed_tterm.

(* Substituting a closed term for [x] within a term where [x] is the sole free
   variable yields a closed term. *)

Lemma closed_subst_sterm:
  forall t1 t2,
  well_scoped_lift lift_sterm 1 t1 ->
  closed_sterm t2 ->
  closed_sterm (subst_sterm t2 0 t1).
Proof.
  unfold closed_sterm, closed, well_scoped_lift. intros.
  rewrite lift_subst_2_sterm; eauto with f_equal.
Qed.

Hint Resolve closed_subst_sterm.

Lemma well_scoped_subst_sterm:
  forall t1 t2 k,
  well_scoped_lift lift_sterm (S k) t1 ->
  well_scoped_lift lift_sterm k t2 ->
  well_scoped_lift lift_sterm k (subst_sterm t2 0 t1).
Proof.
  unfold well_scoped_lift. intros.
  rewrite lift_subst_2_sterm; eauto.
  f_equal; eauto.
Qed.

Hint Resolve well_scoped_subst_sterm : ws.

Lemma well_scoped_subst_tterm:
  forall t1 t2 k,
  well_scoped_lift lift_tterm (S k) t1 ->
  well_scoped_lift lift_tterm k t2 ->
  well_scoped_lift lift_tterm k (subst_tterm t2 0 t1).
Proof.
  unfold well_scoped_lift. intros.
  rewrite lift_subst_2_tterm; eauto.
  f_equal; eauto.
Qed.

Hint Resolve well_scoped_subst_tterm : ws.

(* TEMPORARY
Lemma well_scoped_subst_tterm_1:
  forall t1 t2 k,
  well_scoped_lift lift_tterm (S k) t1 ->
  closed_tterm t2 ->
  well_scoped_lift lift_tterm k (subst_tterm t2 0 t1).
Proof.
  unfold well_scoped_lift. intros.
  rewrite lift_subst_2_tterm; eauto.
  f_equal. rewrite lift_closed_tterm; eauto. eauto.
Qed. *)

Lemma closed_subst_tterm:
  forall t1 t2,
  well_scoped_lift lift_tterm 1 t1 ->
  closed_tterm t2 ->
  closed_tterm (subst_tterm t2 0 t1).
Proof.
  unfold closed_tterm, closed, well_scoped_lift. intros.
  rewrite lift_subst_2_tterm; eauto with f_equal.
Qed.

Hint Resolve closed_subst_tterm.

(* TEMPORARY
Lemma closed_subst_tterm_2:
  forall t t1 t2,
  well_scoped_lift lift_tterm 2 t ->
  closed_tterm t1 ->
  closed_tterm t2 ->
  closed_tterm (subst_tterm t1 0 (subst_tterm (lift_tterm 0 t2) 0 t)).
Proof.
  intros. eapply closed_subst_tterm; eauto. eapply well_scoped_subst_tterm; eauto.
Qed.

Hint Resolve closed_subst_tterm_2.
*)

(* ------------------------------------------------------------------------- *)

(* Closedness is contagious -- this time in the target language. *)

Lemma well_scoped_up_TAbs:
  forall t k,
  well_scoped_lift lift_tterm (S k) t ->
  well_scoped_lift lift_tterm k (TAbs t).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_TApp:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k t1 ->
  well_scoped_lift lift_tterm k t2 ->
  well_scoped_lift lift_tterm k (TApp t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_TPair:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k t1 ->
  well_scoped_lift lift_tterm k t2 ->
  well_scoped_lift lift_tterm k (TPair t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_TLetPair:
  forall t1 t2 k,
  well_scoped_lift lift_tterm k t1 ->
  well_scoped_lift lift_tterm (S (S k)) t2 ->
  well_scoped_lift lift_tterm k (TLetPair t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_TLetPair_0:
  forall t k,
  well_scoped_lift lift_tterm k t ->
  well_scoped_lift lift_tterm k (TLetPair t (TVar 0)).
Proof.
  closed_contagious.
Qed.

Lemma well_scoped_up_TLetPair_1:
  forall t k,
  well_scoped_lift lift_tterm k t ->
  well_scoped_lift lift_tterm k (TLetPair t (TVar 1)).
Proof.
  closed_contagious.
Qed.

Hint Resolve well_scoped_up_TAbs well_scoped_up_TApp
well_scoped_up_TPair well_scoped_up_TLetPair well_scoped_up_TLetPair_0
well_scoped_up_TLetPair_1.

Lemma closed_up_TApp:
  forall t1 t2,
  closed_tterm t1 ->
  closed_tterm t2 ->
  closed_tterm (TApp t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_TPair:
  forall t1 t2,
  closed_tterm t1 ->
  closed_tterm t2 ->
  closed_tterm (TPair t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_TLetPair:
  forall t1 t2,
  closed_tterm t1 ->
  well_scoped_lift lift_tterm 2 t2 ->
  closed_tterm (TLetPair t1 t2).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_TLetPair_0:
  forall t,
  closed_tterm t ->
  closed_tterm (TLetPair t (TVar 0)).
Proof.
  closed_contagious.
Qed.

Lemma closed_up_TLetPair_1:
  forall t,
  closed_tterm t ->
  closed_tterm (TLetPair t (TVar 1)).
Proof.
  closed_contagious.
Qed.

Hint Resolve closed_up_TApp closed_up_TPair closed_up_TLetPair
closed_up_TLetPair_0 closed_up_TLetPair_1.

(* ------------------------------------------------------------------------- *)

(* The encoding of a closed store is closed. *)

Lemma encode_address_get_closed:
  forall l,
  closed_tterm (encode_address_get l).
Proof.
  unfold closed_tterm, closed, well_scoped_lift.
  eauto using lift_encode_address_get.
Qed.

Lemma encode_address_set_closed:
  forall l,
  closed_tterm (encode_address_set l).
Proof.
  unfold closed_tterm, closed, well_scoped_lift.
  eauto using lift_encode_address_set.
Qed.

Lemma encode_address_closed:
  forall l,
  closed_tterm (encode_address l).
Proof.
  unfold closed_tterm, closed, well_scoped_lift.
  eauto using lift_encode_address.
Qed.

Lemma encode_loc_closed:
  forall l,
  closed_tterm (encode_loc l).
Proof.
  unfold encode_loc.
  unfold closed_tterm, closed, well_scoped_lift.
  simpl; intros.
  rewrite lift_encode_address_get.
  rewrite lift_encode_address_set.
  reflexivity.
Qed.

Hint Resolve encode_loc_closed encode_address_get_closed
encode_address_set_closed encode_address_closed.

Lemma encode_heap_closed:
  forall s,
  closed_store s ->
  closed_tterm (encode_heap s).
Proof.
  unfold closed_store. induction s; simpl; eauto 8.
Qed.

Hint Resolve encode_heap_closed.

Lemma encode_store_closed:
  forall s,
  closed_store s ->
  closed_tterm (encode_store s).
Proof.
  unfold encode_store. eauto.
Qed.

Hint Resolve encode_store_closed.

(* ------------------------------------------------------------------------- *)

(* A closed store can be obtained by concatenating two closed stores, by
   creating a singleton store out of a closed term, etc. *)

Lemma closed_store_append:
  forall s1 s2,
  closed_store s1 ->
  closed_store s2 ->
  closed_store (s1 ++ s2).
Proof.
  unfold closed_store. intros. forwards: in_app_or; intuition eauto.
Qed.

Lemma closed_store_singleton:
  forall t,
  closed_sterm t ->
  closed_store (t :: nil).
Proof.
  unfold closed_store; simpl; intuition (subst; eauto).
Qed.

Lemma closed_store_nil:
  closed_store nil.
Proof.
  unfold closed_store. intros. false.
Qed.

Lemma closed_store_cons:
  forall t s,
  closed_sterm t ->
  closed_store s ->
  closed_store (t :: s).
Proof.
  unfold closed_store; simpl; intuition (subst; eauto).
Qed.

Lemma closed_store_cons_invert:
  forall t s,
  closed_store (t :: s) ->
  closed_sterm t /\ closed_store s.
Proof.
  unfold closed_store;simpl; intuition eauto.
Qed.

Lemma closed_skipn:
  forall n s,
  closed_store s ->
  closed_store (skipn n s).
Proof.
  induction n; destruct s; simpl; intros; eauto. 
  forwards [ ? ? ]: closed_store_cons_invert; eauto.
Qed.

Lemma closed_nth:
  forall n s,
  closed_store s ->
  closed_sterm (nth n s SUnit).
Proof.
  induction n; destruct s; simpl; intros; eauto;
  forwards [ ? ? ]: closed_store_cons_invert; eauto.
Qed.

Hint Resolve closed_store_append closed_store_singleton
closed_store_nil closed_store_cons closed_skipn closed_nth.

(* ------------------------------------------------------------------------- *)

(* Reading a closed store yields a closed term. Updating a closed store with
   a closed term yields a closed store. *)

Lemma closed_nth_error:
  forall s l t,
  nth_error s l = value t ->
  closed_store s ->
  closed_sterm t.
Proof.
  eauto using (@forall_nth_error _ closed_sterm).
Qed.

Lemma closed_set_nth_error:
  forall s1 l t s2,
  set_nth_error s1 l t = value s2 ->
  closed_store s1 ->
  closed_sterm t ->
  closed_store s2.
Proof.
  unfold closed_store. intros. eauto using (@forall_set_nth_error _ closed_sterm).
Qed.

Hint Resolve closed_nth_error closed_set_nth_error.

(* ------------------------------------------------------------------------- *)

(* Reduction in the source language preserves closed configurations. *)

Lemma ssimpl_well_scoped:
  forall t1 t2,
  ssimpl t1 t2 ->
  forall k,
  well_scoped_lift lift_sterm k t1 ->
  well_scoped_lift lift_sterm k t2.
Proof.
  induction 1; intros; closed_hereditary; eauto 3 with ws.
Qed.

Lemma ssimpl_closed:
  forall t1 t2,
  ssimpl t1 t2 ->
  closed_sterm t1 ->
  closed_sterm t2.
Proof.
  unfold closed_sterm, closed; eauto using ssimpl_well_scoped.
Qed.

Lemma sred_closed:
  forall s1 E1 t1 s2 E2 t2,
  sred s1 E1 t1 s2 E2 t2 ->
  closed_conf s1 E1 t1 ->
  closed_conf s2 E2 t2.
Proof.
  induction 1; introv [ ? [ ? ? ]]; closed_hereditary; intuition eauto using ssimpl_closed.
Qed.

(* ------------------------------------------------------------------------- *)

(* Reduction in the target language preserves closed terms. *)

Lemma tred_well_scoped:
  forall t1 t2,
  tred t1 t2 ->
  forall k,
  well_scoped_lift lift_tterm k t1 ->
  well_scoped_lift lift_tterm k t2.
Proof.
  induction 1; intros; closed_hereditary; eauto 3 with ws.
  eapply well_scoped_subst_tterm; eauto.
  eapply well_scoped_subst_tterm; eauto.
  eauto using well_scoped_lift_lift, lift_lift_tterm.
Qed.

Lemma tred_closed:
  forall t1 t2,
  tred t1 t2 ->
  closed_tterm t1 ->
  closed_tterm t2.
Proof.
  unfold closed_tterm, closed; eauto using tred_well_scoped.
Qed.

Lemma tredstar_closed:
  forall t1 t2,
  tredstar t1 t2 ->
  closed_tterm t1 ->
  closed_tterm t2.
Proof.
  induction 1; eauto using tred_closed.
Qed.

