Set Implicit Arguments.
Require Import MyTactics.
Require Import MyLists.
Require Import DeBruijn.
Require Import SourceSyntax.
Require Import TargetSyntax.
Require Import EncodingBricks.

(* This is the definition of the encoding, as per the paper. *)

(* Here, we are looking at the encoding of untyped terms. There
   are no type annotations in the source or target languages. *)

(* ------------------------------------------------------------------------- *)

(* The encoding of memory locations. *)

(* The encoding of memory locations is in several layers. First, the encoding
   of a raw address [l] is a pair of [get] and [set] functions, whose code is
   given by an inductive function of [l]. (There is a bit of painful manual
   work here; these definitions would have to be adapted if the FORK files
   were modified.) Second, the encoding of a memory location is (again) a pair
   of [get] and [set] functions, which are obtained by applying [__array_read]
   and [__array_write] to (the encoding of) the raw address. *)

(* For more details about this, see the FORK file [Arrays.f]. In short, the
   basic reason why we have two layers of [get] and [set] functions is that
   the functions in the lower layer are designed to access only the last
   element of the array (more precisely, its tail, when the array is viewed as
   a list), while the functions in the upper layer are designed to access an
   element at an arbitrary index. *)

Fixpoint encode_address_get (l : nat) : tterm :=
  match l with
  | 0 =>
      __zero_get
  | S l =>
      (* this is [TApp __successor_get (encode_address_get l)], beta-reduced *)
      (* \h. let (_, tail) = get h in tail *)
      TAbs (TLetPair (TApp (encode_address_get l) (TVar 0)) (TVar 0))
  end.

Fixpoint encode_address_set (l : nat) : tterm :=
  match l with
  | 0 =>
      __zero_set
  | S l =>
      (* this is [TApp __successor_set (encode_address_get l) (encode_address_set l)], beta-reduced *)
      (* \h. \x. let (data, _) = get h in set h (data, x) *)
      TAbs (TAbs (
        TLetPair (TApp (encode_address_get l) (TVar 1)) (
          TApp (TApp (encode_address_set l) (TVar 3)) (TPair (TVar 1) (TVar 2))
        )
      ))
  end.

Definition encode_address (l : nat) : tterm :=
  TPair (encode_address_get l) (encode_address_set l).

Definition encode_loc (l : nat) : tterm :=
  TPair
    (TAbs (TApp (TApp __array_read  (TVar 0)) (encode_address l)))
    (TAbs (TApp (TApp __array_write (TVar 0)) (encode_address l))).

(* ------------------------------------------------------------------------- *)

(* The encoding of terms. *)

(* The encoding is as in the paper, stripped of any type annotations. It is
   the identity over pure terms. It translates the monadic constants to
   appropriate terms, extracted from FORK. It translates memory locations as
   per the above definitions. *)

Fixpoint encode (t : sterm) : tterm :=
  match t with
  | SVar x =>
      TVar x
  | SAbs t =>
      TAbs (encode t)
  | SApp t1 t2 =>
      TApp (encode t1) (encode t2)
  | SUnit =>
      TUnit
  | SPair t1 t2 =>
      TPair (encode t1) (encode t2)
  | SProjLeft t =>
      TLetPair (encode t) (TVar 1)
  | SProjRight t =>
      TLetPair (encode t) (TVar 0)
  | SReturn =>
      __return (* defined in EncodingBricks *)
  | SBind =>
      bind (* defined in EncodingBricks *)
  | SNew =>
      new (* defined in EncodingBricks *)
  | SRead =>
      read (* defined in EncodingBricks *)
  | SWrite =>
      write (* defined in EncodingBricks *)
  | SLoc l =>
      encode_loc l
  end
.

(* ------------------------------------------------------------------------- *)

(* The encoding of stores. *)

(* A store is first encoded as a heap (that is, a raw data array) using nested
   pairs. *)

Fixpoint encode_heap (s : store) :=
  match s with
  | nil =>
      TUnit
  | t :: s =>
      TPair (encode t) (encode_heap s)
  end.

(* Then, a store can be encoded as an array (that is, a pair of a heap and an
   end address). *)

Definition encode_store (s : store) :=
  TPair (encode_heap s) (encode_address (length s)).

(* The term that is presented as the encoding of the empty store in the FORK
   code is indeed the encoding of the empty store. *)

Lemma encode_store_nil:
  encode_store nil = store_empty.
Proof.
  unfold encode_store. simpl. reflexivity.
Qed.

(* ------------------------------------------------------------------------- *)

(* The encoding of evaluation contexts. *)

(* The expression [bind t1 t2 s] first evaluates the expression [t1] in the
   store [s]. To do so, it reduces to a term where the application [t1 s]
   appears in an evaluation context that depends on [t2]. We refer to this
   term as [ebind (TApp t1 s) t2]. The exact definition of [ebind] is not
   particularly enlightening -- it is obtained simply by reducing [bind t1 t2
   s] step by step until the term is in the desired form. *)

Definition ebind t1 t2 :=
  TLetPair t1 (
    TLet (TVar 1) (
      TLet (TVar 1) (
        TLet (TApp (TApp t2 (TVar 1)) (TVar 0)) (
          TApp backward (TVar 0)
        )
      )
    )
  ).

(* An evaluation context is a stack of [bind] combinators that have been
   partially executed, so their natural encoding is [ebind]. *)

Fixpoint encode_ectx E u :=
  match E with
  | EHole =>
      u
  | EBind E t =>
      encode_ectx E (ebind u (encode t))
  end.

(* ------------------------------------------------------------------------- *)

(* The encoding of configurations. *)

(* The source language does not have a [run] operation. We assume that the
   main program has type [M tau] for some type [tau], and it is initially run
   in an empty store. *)

(* In order to encode a configuration of a store [s], an evaluation context
   [E], and a main program [t], we encode the term [t] -- this yields a
   function that expects a store as an argument -- and we apply this function
   to the encoding of the store [s]. Then, we place this term within the
   encoding of the evauation context [E]. *)

Notation encode_conf s E t :=
  (encode_ectx E (TApp (encode t) (encode_store s))).

(* ------------------------------------------------------------------------- *)

(* The encoding of an address is a closed term. *)

Lemma lift_encode_address_get:
  forall l k,
  lift_tterm k (encode_address_get l) = encode_address_get l.
Proof.
  induction l; simpl; intros; try rewrite IHl; eauto with f_equal.
Qed.

Lemma lift_encode_address_set:
  forall l k,
  lift_tterm k (encode_address_set l) = encode_address_set l.
Proof.
  induction l; simpl; intros; try rewrite IHl; eauto using lift_encode_address_get with f_equal.
Qed.

Lemma lift_encode_address:
  forall l k,
  lift_tterm k (encode_address l) = encode_address l.
Proof.
  unfold encode_address. simpl.
  eauto using lift_encode_address_get, lift_encode_address_set with f_equal.
Qed.

Lemma subst_encode_address_get:
  forall l k u,
  subst_tterm u k (encode_address_get l) = encode_address_get l.
Proof.
  induction l; simpl; intros; try rewrite IHl; eauto with f_equal.
Qed.

Lemma subst_encode_address_set:
  forall l k u,
  subst_tterm u k (encode_address_set l) = encode_address_set l.
Proof.
  induction l; simpl; intros; try rewrite IHl; eauto using subst_encode_address_get with f_equal.
Qed.

Lemma subst_encode_address:
  forall l k u,
  subst_tterm u k (encode_address l) = encode_address l.
Proof.
  unfold encode_address. simpl.
  eauto using subst_encode_address_get, subst_encode_address_set with f_equal.
Qed.

(* ------------------------------------------------------------------------- *)

(* The encoding function commutes with lifting. *)

Lemma encode_lift:
  forall t k,
  encode (lift_sterm k t) = lift_tterm k (encode t).
Proof.
  induction t; simpl; intros;
  try rewrite lift_encode_address_get;
  try rewrite lift_encode_address_set;
  eauto with f_equal.
Qed.

(* ------------------------------------------------------------------------- *)

(* The encoding of a closed term is closed. *)

Lemma encode_closed:
  forall t,
  closed_sterm t ->
  closed_tterm (encode t).
Proof.
  unfold closed_sterm, closed_tterm, closed, well_scoped_lift.
  intros. rewrite <- encode_lift. f_equal. eauto.
Qed.

Hint Resolve encode_closed.

Lemma lift_encode_closed:
  forall t k,
  closed_sterm t ->
  lift_tterm k (encode t) = encode t.
Proof.
  eauto using lift_closed_tterm.
Qed.

(* ------------------------------------------------------------------------- *)

(* The encoding function commutes with substitution -- it is compositional. *)

Lemma encode_subst:
  forall t u k,
  encode (subst_sterm u k t) = subst_tterm (encode u) k (encode t).
Proof.
  induction t; intros; eauto; simpl;
  try rewrite <- encode_lift; (* SAbs *)
  try rewrite subst_encode_address_get; (* SLoc *)
  try rewrite subst_encode_address_set; (* SLoc *)
  f_equal; eauto.
  (* SVar *)
  unfold subst_var; by_cases; auto.
Qed.

Lemma subst_encode_closed:
  forall u t k,
  closed_sterm t ->
  subst_tterm u k (encode t) = encode t.
Proof.
  eauto using subst_closed_tterm.
Qed.

