Set Implicit Arguments.
Require Export List.
Require Import MyTactics.

(* ------------------------------------------------------------------------- *)

(* Setting the n-th element of a list. *)

Fixpoint set_nth_error (A : Type) (l : list A) (n : nat) (v : A) : Exc (list A) :=
  match n, l with
  | O, _ :: l =>
      value (v :: l)
  | S n, x :: l =>
      match set_nth_error l n v with
      | value l => value (x :: l)
      | error => error
      end
  | _, _ => error
  end.

(* ------------------------------------------------------------------------- *)

(* Properties. *)

Lemma forall_nth_error:
  forall (A : Type) (P : A -> Prop) xs k x,
  nth_error xs k = value x ->
  (forall x, In x xs -> P x) ->
  P x.
Proof.
  induction xs; destruct k; simpl; introv h1 h2; dependent destruction h1; eauto.
Qed.

Lemma forall_set_nth_error:
  forall (A : Type) (P : A -> Prop) xs1 k x xs2,
  set_nth_error xs1 k x = value xs2 ->
  (forall x, In x xs1 -> P x) ->
  P x ->
  (forall x, In x xs2 -> P x).
Proof.
  induction xs1; destruct k; simpl; introv h1 h2 hx; dependent destruction h1; simpl; introv i.
  destruct i; subst; eauto.
  case_eq (set_nth_error xs1 k x); introv eq.
    (* subcase: value *)
    match goal with h: _ = value _ |- _ => rewrite eq in h; injection h; clear h; intro; subst end.
    simpl in i. intuition eauto.
    (* subcase: error *)
    match goal with h: _ = value _ |- _ => rewrite eq in h; discriminate h end.
Qed.

Lemma nth_error_length:
  forall (A : Type) n (s : list A) x,
  nth_error s n = value x ->
  n < length s.
Proof.
  induction n; destruct s; simpl; introv eq; dependent destruction eq; eauto.
  forwards: IHn; eauto.
Qed.

Lemma set_nth_error_length:
  forall (A : Type) n (s1 s2 : list A) x,
  set_nth_error s1 n x = value s2 ->
  n < length s1.
Proof.
  induction n; destruct s1; simpl; intros; try discriminate; eauto.
  case_eq (set_nth_error s1 n x); introv h; rewrite h in *; try discriminate.
  forwards: IHn; eauto.
Qed.

Lemma set_nth_error_preserves_length:
  forall (A : Type) (s1 s2 : list A) n x,
  set_nth_error s1 n x = value s2 ->
  length s1 = length s2.
Proof.
  induction s1; destruct n; simpl; intros; try discriminate; unfold value in *.
  dependent destruction H. eauto.
  case_eq (set_nth_error s1 n x); introv h; rewrite h in *; try discriminate.
  dependent destruction H.
  forwards: IHs1; simpl; eauto.
Qed.

Hint Resolve nth_error_length set_nth_error_length.

Lemma firstn_nth:
  forall (A : Type) n (s tail : list A) d,
  n < length s ->
  firstn n s ++ nth n s d :: tail = firstn (S n) s ++ tail.
Proof.
  induction n; destruct s; simpl; intros; try (false; omega); eauto.
  f_equal. rewrite IHn; eauto.
Qed.

Lemma firstn_x_skipn:
  forall (A : Type) n (s1 : list A) x s2,
  set_nth_error s1 n x = value s2 ->
  firstn n s1 ++ x :: skipn (S n) s1 = s2.
Proof.
  induction n; destruct s1; simpl; introv eq; try discriminate; unfold value in *.
  congruence.
  case_eq (set_nth_error s1 n x); introv h; rewrite h in *; try discriminate.
  dependent destruction eq. f_equal.
  forwards: IHn; eauto.
Qed.

Lemma firstn_length:
  forall (A : Type) (s : list A),
  firstn (length s) s = s.
Proof.
  induction s; simpl; f_equal; eauto.
Qed.

Lemma nth_error_nth:
  forall (A : Type) n (s : list A) d x,
  nth_error s n = value x ->
  nth n s d = x.
Proof.
  induction n; destruct s; simpl; intros; try discriminate.
  unfold value in *. congruence.
  eauto.
Qed.

