Require Import MyLists.
Require Import MyTactics.
Require Import DeBruijn.
Require Import SourceSyntax.

(* Following Moggi, the operational semantics of the source language is in two
   layers. The pure layer defines a simplification relation, which is just the
   usual notion of beta-reduction of lambda-terms. Beta-reduction is permitted
   everywhere except under lambda-abstractions. The impure layer defines the
   reduction of configurations, where configurations include a store, an
   evaluation context, and a term. An evaluation context [E] is either a hole
   or of the form E[bind x = [] in t]. *)

Inductive ssimpl : sterm -> sterm -> Prop :=
| SSimplBeta:
    forall t1 t2,
    ssimpl
      (SApp (SAbs t1) t2)
      (subst_sterm t2 0 t1)
| SSimplProjLeft:
    forall t1 t2,
    ssimpl
      (SProjLeft (SPair t1 t2))
      t1
| SSimplProjRight:
    forall t1 t2,
    ssimpl
      (SProjRight (SPair t1 t2))
      t2
| SSimplAbsContext:
    forall t1 t2,
    ssimpl t1 t2 ->
    ssimpl (SAbs t1) (SAbs t2)
| SSimplAppLeftContext:
    forall t1 t2 t,
    ssimpl t1 t2 ->
    ssimpl (SApp t1 t) (SApp t2 t)
| SSimplAppRightContext:
    forall t1 t2 t,
    ssimpl t1 t2 ->
    ssimpl (SApp t t1) (SApp t t2)
| SSimplPairLeftContext:
    forall t1 t2 t,
    ssimpl t1 t2 ->
    ssimpl (SPair t1 t) (SPair t2 t)
| SSimplPairRightContext:
    forall t1 t2 t,
    ssimpl t1 t2 ->
    ssimpl (SPair t t1) (SPair t t2)
| SSimplProjLeftContext:
    forall t1 t2,
    ssimpl t1 t2 ->
    ssimpl (SProjLeft t1) (SProjLeft t2)
| SSimplProjRightContext:
    forall t1 t2,
    ssimpl t1 t2 ->
    ssimpl (SProjRight t1) (SProjRight t2)
.

Inductive sred : store -> ectx -> sterm -> store -> ectx -> sterm -> Prop :=
| SRedSimpl:
    forall s E t1 t2,
    ssimpl t1 t2 ->
    sred
      s E t1
      s E t2
| SRedBind:
    forall s E t1 t2,
    sred
      s E (SApp SBind (SPair t1 t2))
      s (EBind E t2) t1
| SRedReturn:
    forall s E t1 t2,
    sred
      s (EBind E t2) (SApp SReturn t1)
      s E (SApp t2 t1)
| SRedNew:
    forall s E t,
    sred
      s E (SApp SNew t)
      (s ++ t :: nil) E (SApp SReturn (SLoc (length s)))
| SRedRead:
    forall s E l t,
    nth_error s l = value t ->
    sred
      s E (SApp SRead (SLoc l))
      s E (SApp SReturn t)
| SRedWrite:
    forall s1 s2 E l t,
    set_nth_error s1 l t = value s2 ->
    sred
      s1 E (SApp SWrite (SPair (SLoc l) t))
      s2 E (SApp SReturn SUnit)
.

Inductive sredstar : store -> ectx -> sterm -> store -> ectx -> sterm -> Prop :=
| SRedStarZero:
    forall s E t,
    sredstar s E t s E t
| SRedStarMore:
    forall s1 E1 t1 s2 E2 t2 s3 E3 t3,
    sred s1 E1 t1 s2 E2 t2 ->
    sredstar s2 E2 t2 s3 E3 t3 ->
    sredstar s1 E1 t1 s3 E3 t3.

(* A little lemma: simplification preserves the form [SApp SReturn _]. *)

Lemma ssimpl_return:
  forall t1 u2,
  ssimpl (SApp SReturn t1) u2 ->
  exists t2,
  u2 = SApp SReturn t2 /\
  ssimpl t1 t2.
Proof.
  introv h. dependent destruction h.
  dependent destruction h.
  intuition eauto.
Qed.

