(**************************************************************************)
(*                                                                        *)
(*  FORK                                                                  *)
(*                                                                        *)
(*  Author:  Franois Pottier, INRIA Paris-Rocquencourt                   *)
(*  Version: 20101109                                                     *)
(*                                                                        *)
(*  The copyright to this code is held by Institut National de Recherche  *)
(*  en Informatique et en Automatique (INRIA). All rights reserved. This  *)
(*  file is distributed under the license CeCILL-C (see file LICENSE).    *)
(*                                                                        *)
(**************************************************************************)

open Atom
open Kinds

(* The types in FORK are pure lambda-terms with constants. *)

type constant =
  | TyUnit
  | TyArrow
  | TyPair
  | TyForall of kind
  | TyExists of kind

type ty =
  | TyVar of atom
  | TyConstant of constant
  | TyAbs of atom * kind * locty
  | TyApp of locty * locty
  | TyConstraint of locty * kind

and locty =
    Error.location * ty

(* Types can be (recursively) defined by the user. A type signature
   maps atoms to types. *)

type signature =
    (kind * locty) AtomMap.t

(* A kind environment maps type variables to kinds. *)

type env =
    kind AtomMap.t

(* ------------------------------------------------------------------------- *)

(* The kinds of the type constants. *)

let kind_of_constant = function
  | TyUnit ->
      KStar
  | TyArrow
  | TyPair ->
      KArrow (KLater KStar, KArrow (KLater KStar, KStar)) (* TEMPORARY!! experimental *)
  | TyForall kind
  | TyExists kind ->
      KArrow (KArrow (kind, KStar), KStar)
      (* TEMPORARY should be polymorphic in a "later/star" continuation kind? *)

(* ------------------------------------------------------------------------- *)

(* Kind checking. *)

(* Because every type variable is annotated with its kind, kind checking is
   a simple bottom-up process. *)

let rec infer env (loc, ty) =
  match ty with
  | TyVar a ->
      AtomMap.find a env
  | TyConstant c ->
      kind_of_constant c
  | TyAbs (a, kind, ty) ->
      let env = AtomMap.add a kind env in
      KArrow (kind, infer env ty)
  | TyApp (ty1, ty2) ->
      kapp [loc] (infer env ty1) (infer env ty2)
  | TyConstraint (ty, declared_kind) ->
      check env loc ty declared_kind

and check env loc ty declared_kind =
  let inferred_kind = infer env ty in
  if is_subkind inferred_kind declared_kind then
    declared_kind
  else
    raise (KindChecking ([loc], Printf.sprintf
      "This kind annotation is invalid.\n\
       This type has kind:\n\
       %s"
      (print_kind inferred_kind)
    ))

(* ------------------------------------------------------------------------- *)

(* Dependency analysis among the toplevel type definitions. *)

(* We build a graph of the dependencies among toplevel type definitions and
   determine its strongly connected components. *)

let scc types =

  let module G = struct

    type node =
      atom
	
    let n, index =
      AtomMap.index types

    let successors f a =
      let rec iter env (_, ty) =
	match ty with
	| TyVar a ->
	    if not (AtomSet.mem a env) then
	      f a
	| TyConstant _ ->
	    ()
	| TyAbs (a, _, ty) ->
	    iter (AtomSet.add a env) ty
	| TyApp (ty1, ty2) ->
	    iter env ty1;
	    iter env ty2
	| TyConstraint (ty, _) ->
	    iter env ty
      in
      let _, ty = AtomMap.find a types in
      iter AtomSet.empty ty

    let iter f =
      AtomMap.iter (fun a _ -> f a) types

  end in

  let module T = Tarjan.Run (G) in

  fun a -> T.scc (T.representative a)

(* ------------------------------------------------------------------------- *)

(* Kind checking the toplevel type definitions. *)

(* The type definitions found at the top level are considered recursive. For
   this reason, they are annotated with a kind. Checking these definitions
   involves matching the inferred kind against the declared kind. *)

(* As we check a definition, the environment is weakened by inserting a bullet
   in front of every assumption that belongs in the same strongly connected
   component. *)

(* Within a single strongly connected component, it would be possible to adopt
   a somewhat flexible rule -- roughly speaking, it is sufficient to insert
   enough bullets to break every cycle -- but there is in general no canonical
   way of doing so. *)

let weaken scc a env =
  List.fold_left (fun env b ->
    let declared_kind = AtomMap.find b env in
    AtomMap.add b (KLater declared_kind) env
  ) env (scc a)

let check types =
  let scc =
    scc types
  in
  let env =
    AtomMap.map (fun (declared_kind, _) -> declared_kind) types
  in
  AtomMap.iter (fun a (declared_kind, ty) ->
    try
      let env = weaken scc a env in
      let loc, _ = ty in
      let (_ : kind) = check env loc ty declared_kind in
      ()
    with KindChecking (locs, msg) ->
      Error.signal locs msg
  ) types;
  Error.signaled();
  env

